import dataclasses


@dataclasses.dataclass(frozen=True)
class KeyTweets:
    """Dataclass storing key convention for tweets datasets"""

    x_ngt: str = "x_ngt"
    y_ngt: str = "y_ngt"
    x_ngt_rounded: str = "x_ngt_rounded"
    y_ngt_rounded: str = "y_ngt_rounded"
    time: str = "created_at"
    time_radar: str = "time_radar"
    time_stations: str = "time_half"
    time_radar_int: str = "time_radar_int"
    latitude: str = "latitude"
    latitude_stations: str = "station_latitude"
    longitude: str = "longitude"
    longitude_stations: str = "station_longitude"
    tp_radar: str = "tp_mm_radar"
    station_tp: str = "station_tp_mm"
    distance_stations: str = "station_distance_km"



import contextlib
import logging
import pathlib
import typing as t
import warnings

import a2.plotting.analysis

try:
    import mantik
except ImportError as e:
    logging.warn("Mantik couldn't be imported\n%s", e)
try:
    import mlflow
except ImportError as e:
    logging.warn("mlflow couldn't be imported\n%s", e)


def initialize_mantik():
    """
    Attemps to initialize mantik, throws exception if fails
    """
    try:
        mantik.init_tracking()
    except Exception as e:
        warnings.warn(f"{e}\nCannot initialize mantik!")


def catch_mantik_exceptions(func):
    def wrapper_catch_exception(*args, **kwargs):
        result = None
        try:
            if not args[0].ignore:
                result = func(*args, **kwargs)
        except mlflow.exceptions.MlflowException as e:
            logging.info(f"Ignoring mlflow exception:\n{e}")
        except Exception as exc:
            raise ValueError("Tracking error!") from exc
        return result

    return wrapper_catch_exception


class Tracker:
    def __init__(self, ignore=False) -> None:
        self.ignore = ignore
        if not self.ignore:
            initialize_mantik()

    @catch_mantik_exceptions
    def log_param(self, name, value):
        mlflow.log_param(name, value)

    @catch_mantik_exceptions
    def log_params(self, params, **kwargs):
        mlflow.log_params(params)

    @catch_mantik_exceptions
    def log_metric(self, *args, **kwargs):
        mlflow.log_metric(*args, **kwargs)

    @catch_mantik_exceptions
    def log_metrics(self, *args, **kwargs):
        mlflow.log_metrics(*args, **kwargs)

    @contextlib.contextmanager
    def start_run(self, *args, **kwargs):
        if not self.ignore:
            yield mlflow.start_run(*args, **kwargs)
        else:
            yield None

    @catch_mantik_exceptions
    def end_run(self, *args, **kwargs):
        mlflow.end_run(*args, **kwargs)

    @catch_mantik_exceptions
    def log_artifact(self, *args, **kwargs):
        mlflow.log_artifact(*args, **kwargs)

    @catch_mantik_exceptions
    def active_run(self, *args, **kwargs):
        return mlflow.active_run(*args, **kwargs)

    @catch_mantik_exceptions
    def local_file_uri_to_path(self, *args, **kwargs):
        return mlflow.utils.file_utils.local_file_uri_to_path(*args, **kwargs)

    # @catch_mantik_exceptions
    def set_tracking_uri(self, *args, **kwargs):
        mlflow.set_tracking_uri(*args, **kwargs)

    @catch_mantik_exceptions
    def get_tracking_uri(self, *args, **kwargs):
        return mlflow.get_tracking_uri(*args, **kwargs)

    @catch_mantik_exceptions
    def create_experiment(self, name, **kwargs):
        experiment_id = self.get_experiment_by_name(name)
        print(f"{experiment_id=}")
        logging.info(f"{not experiment_id=}")
        if not experiment_id:
            experiment_id = mlflow.create_experiment(name, **kwargs)
            logging.info(f"create new: {experiment_id}")
        if experiment_id is not None and not isinstance(experiment_id, str):
            experiment_id = experiment_id.experiment_id
        return experiment_id

    @catch_mantik_exceptions
    def get_experiment_by_name(self, *args, **kwargs):
        return mlflow.get_experiment_by_name(*args, **kwargs)


def log_metric_classification_report(
    tracker: Tracker,
    truth: t.Sequence,
    predictions: t.Sequence,
    step: int | None = 1,
    label: str = "raining",
    filename_confusion_matrix: pathlib.Path | str = "confusion_matrix.pdf",
    font_size: int = 14,
):
    """
    Compute f1 score and logs results to mlflow

    Parameters:
    ----------
    truth: True labels
    predictions: Predicted labels
    prediction_probabilities: Prediction probability for both labels, shape = [n_tests, 2]
    Step: Current training stop (epoch)
    label: Label name of the classification

    Returns
    -------
    """
    initialize_mantik()
    classification_report = a2.plotting.analysis.check_prediction(
        truth=truth,
        prediction=predictions,
        filename=filename_confusion_matrix,
        output_dict=True,
        label=label,
        font_size=font_size,
    )
    logging.info(classification_report)
    log_classification_report(tracker, classification_report, step, label)
    tracker.log_artifact(filename_confusion_matrix)


def log_classification_report(tracker, classification_report, step, label):
    initialize_mantik()
    tracker.log_metric(
        key=f"eval_f1_{label}",
        value=classification_report[label]["f1-score"],
        step=step,
    )
    tracker.log_metric(
        key=f"eval_f1_not_{label}",
        value=classification_report[f"not {label}"]["f1-score"],
        step=step,
    )
    tracker.log_metric(
        key="weighted average f1-score",
        value=classification_report["weighted avg"]["f1-score"],
        step=step,
    )
    tracker.log_metric(
        key="macro average f1-score",
        value=classification_report["macro avg"]["f1-score"],
        step=step,
    )



import ast
import collections
import datetime
import functools
import itertools
import logging
import multiprocessing
import pathlib
import re
import typing as t
import warnings
from collections.abc import Iterable
from functools import wraps
from time import time
from typing import Optional

import numpy as np
import pandas as pd


def timing(f: t.Callable):
    """
    Wrapper that returns execution time and arguments of function.
    """

    @wraps(f)
    def wrap(*args, **kw):
        ts = time()
        result = f(*args, **kw)
        te = time()
        logging.info(f"func:{f.__name__} args:[{kw}] took: {te-ts} sec")
        return result

    return wrap


def get_processes(processes):
    if processes < 0:
        processes = multiprocessing.cpu_count() + processes + 1
    return processes


def parallelize(
    function: t.Callable,
    args_zipped: t.Iterable,
    processes: int = -1,
    single_arg: bool = False,
    kwargs_as_dict: t.Optional[dict] = None,
):
    """
    parallelize function with args provided in zipped format
    ----------
    function: function to parallelize
    args: args of function in zipped format

    Returns
    -------
    function applied to args
    """
    if processes == 1:
        results = []
        if kwargs_as_dict is None:
            kwargs_as_dict = {}
        for args in args_zipped:
            if isinstance(args, str):
                args = (args,)
            results.append(function(*args, **kwargs_as_dict))
        return results
    processes = get_processes(processes)
    print(f"n processes: {processes}")
    if single_arg and kwargs_as_dict is None:
        with multiprocessing.Pool(processes=processes) as pool:
            results = pool.map(function, args_zipped)
    else:
        if single_arg:
            args_zipped = ((arg,) for arg in args_zipped)
        if kwargs_as_dict is not None:
            kwargs_iter = itertools.repeat(kwargs_as_dict)
        else:
            kwargs_iter = itertools.repeat(dict())
        with multiprocessing.Pool(processes=processes) as pool:
            results = starmap_with_kwargs(pool, function, args_zipped, kwargs_iter)
    return results


def get_random_indices(n_sample: int, size_data: int) -> np.ndarray:
    """Draw `n_sample` indices from range(`size_data`)"""
    return np.random.choice(
        range(size_data),
        n_sample if n_sample < size_data else size_data,
        replace=False,
    )


def starmap_with_kwargs(pool, function: t.Callable, args_iter: t.Iterable, kwargs_iter: t.Iterable):
    """Helper function to parallelize functions with args and kwargs"""
    if kwargs_iter is None:
        args_for_starmap = zip(itertools.repeat(function), args_iter)
    else:
        args_for_starmap = zip(itertools.repeat(function), args_iter, kwargs_iter)
    return pool.starmap(apply_args_and_kwargs, args_for_starmap)


def apply_args_and_kwargs(fn: t.Callable, args, kwargs):
    """Helper function to parallelize functions with args and kwargs"""
    return fn(*args, **kwargs)


def evaluate_string(x: object):
    try:
        return ast.literal_eval(str(x))
    except ValueError:
        return x


def chunks(sequence: t.Sequence, n: int) -> t.Iterator:
    """Yield successive `n`-sized chunks from `sequence`."""
    if n == -1:
        n = len(sequence)
    for i in range(0, len(sequence), n):
        yield sequence[slice(i, i + n)]


def round_to_base(x: float, base: float):
    return base * round(x / base)


def round_offset(values: np.ndarray, decimal: int, offset: float) -> np.ndarray:
    values += offset
    values = np.round(values, decimals=decimal)
    values -= offset
    return values


def round_numpy_time_to_base_minutes(time: np.datetime64, base: int = 5) -> np.datetime64:
    tm = datetime.datetime.utcfromtimestamp(time.tolist() / 1e9)
    tm += datetime.timedelta(minutes=base / 2)
    tm -= datetime.timedelta(minutes=tm.minute % base, seconds=tm.second, microseconds=tm.microsecond)
    return np.datetime64(tm)


def round_time_to_base_minutes(time: np.datetime64, base: int = 5):
    tm = pd.to_datetime(time).round(f"{base}min")
    return np.datetime64(tm)


def vetorize_time_to_base_minutes(times: np.ndarray, base: int = 5):
    function = functools.partial(round_numpy_time_to_base_minutes, base=base)
    rounded = np.array(list(map(function, times)))
    return rounded


def _assert_same_type(variable, type_from_variable):
    """Check if `variable` shares type with `type_from_variable`"""
    if isinstance(variable, collections.abc.Iterable):
        if not isinstance(type_from_variable, collections.abc.Iterable):
            raise TypeError(f"Variable {variable=} is iterable but type is not {type_from_variable=}")
        for v, typ in zip(variable, type_from_variable):
            _assert_same_type(v, typ)
    else:
        if not isinstance(variable, type(type_from_variable)):
            raise TypeError(f"Variable {variable=} different type than expected {type_from_variable=}")


def assert_same_type_as(variable: object, type_from_variable: object, alternative: Optional[object] = None):
    """Check if `variable` is same type as `type_from_variable` unless `variable` is `alternative`"""
    if variable is alternative:
        return
    try:
        _assert_same_type(variable, type_from_variable)
    except TypeError:
        raise TypeError(f"Variable {variable} not of expected type {type_from_variable}")


def all_same_type(variable_list: t.Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def assert_shape(variable, shape: t.Tuple, name: Optional[str] = None, ignore_none: bool = True):
    if ignore_none:
        if variable is None:
            return
    if np.shape(variable) != shape:
        raise TypeError(f"{name+': ' if name is not None else ''}{variable=} doesn't have required {shape=}!")


def flatten_list(lis: list[list]) -> list:
    return [item for sublist in lis for item in sublist]


def str_to_delta_time(string: str) -> t.Tuple[float, str]:
    time, units, _ = re.split("([a-zA-Z]+)$", string)
    return float(time), units


def to_nlength_tuple(x: object, n: int = 2) -> tuple:
    if not (isinstance(x, list) or isinstance(x, tuple) or isinstance(x, np.ndarray)):
        return tuple([x] * n)
    elif isinstance(x, list) or isinstance(x, np.ndarray):
        x = tuple(x)
    if len(x) != n:
        raise ValueError(f"{x} doesn't have expected length {n=}")
    return x


def assert_all_same_type(variable_list: Iterable, type_: type):
    for var in variable_list:
        if not isinstance(var, type_):
            raise ValueError(f"{var} not of type {type_.__name__}!")


def validate_array(array: np.ndarray, type_: str = "float", name: str | None = None):
    """Validate numpy based on `type`, e.g. no nan-values"""
    if type_ == "float":
        if np.sum(np.isnan(array.astype(float))) > 0:
            raise ValueError(f"Found nan-values {f'in {name}' if name is not None else ''}!")
    else:
        raise ValueError(f"{type_=} unknown!")


def override_class_method(class_instance, method_name, target_class):
    class_method = getattr(target_class, method_name)

    def new_method(*args, **kwargs):
        return class_method(class_instance, *args, **kwargs)

    setattr(class_instance, method_name, new_method)


def is_in_list_and_remove(to_check, _list):
    is_in = False
    if to_check in _list:
        is_in = True
    if is_in:
        _list.remove(to_check)
    return is_in


def _import_torch(file):
    try:
        import torch as torch  # noqa: E501
    except ModuleNotFoundError:
        warnings.warn(f"Need to install `torch` to use all functionality in {pathlib.Path(file).parent}.")
    else:
        return torch



import a2.utils.utils

torch = a2.utils.utils._import_torch(__file__)


def cuda_available():
    return torch.cuda.is_available()


def available_cuda_devices():
    return [torch.cuda.device(i) for i in range(torch.cuda.device_count())]


def available_cuda_devices_names():
    return [torch.cuda.get_device_name(i) for i in range(torch.cuda.device_count())]



import logging
import os.path
import pathlib

import a2.utils.constants


def save_figure(
    fig: a2.utils.constants.TYPE_MATPLOTLIB_FIGURES, filename: str | pathlib.Path | None = None, dpi: int = 450
) -> None:
    """Save figure to filename"""
    if filename is not None:
        logging.info(f"... saving {filename}")
        folder = os.path.split(filename.__str__())[0]
        if folder:
            os.makedirs(folder, exist_ok=True)
        fig.savefig(filename, bbox_inches="tight", dpi=dpi)
